*************************************************************************************
*														*
*	Bitmap image loader for the EASy68k simulator V 0.95	2007/09/29			*
*														*
*	This code shows how a bitmap can be loaded into EASy68K memory and then		*
*	rendered on the screen. The bitmap can be a 1, 4, 8 or 24 bit colour image	*
*	and is rendered in 24 bit using the correct colours.					*
*														*
*	No checks are made to see if the bitmap will fit on the screen, that the	*
*	width or height is not too large to cause errors or that it will fit into	*
*	the available memory. That said it has worked with all the bitmap files I	*
*	tried so far.											*
*														*
*	After the bitmap has displayed hitting any key will restart the program so	*
*	that another file can be loaded.								*
*														*
*	This version for Sim68K 3.7 or later							*
*														*
*	More 68000 and other projects can be found on my website at ..			*
*														*
*	 http://www.themotionstore.com/leeedavison/index.html					*
*														*
*	.. or ..												*
*														*
*	 http://members.lycos.co.uk/leeedavison/index.html					*
*														*
*	mail : leeedavison@lycos.co.uk								*
*														*
*************************************************************************************

	ORG	$1000

start
	BSR.s		Initialise			* go setup stuff

	BSR.s		open_file			* go open a file
	MOVEQ		#-1,d0			* set test value
	CMP.l		d0,d1				* test the file ID
	BEQ.s		abort_load			* if no file just quit

	BSR		load_file			* load the file
	TST.l		d1				* test the returned flag
	BMI.s		abort_load			* if something was wrong just quit

	BSR		bm_image_header		* get the image header information
	TST.l		d1				* test the returned flag
	BMI.s		abort_load			* if something was wrong just quit

	BSR		display_image		* was all ok so show the image
abort_load
	MOVEQ		#7,d0				* test key waiting
	TRAP		#15

	TST.b		d1				* test the result
	BEQ.s		abort_load			* loop if no key

	MOVEQ		#5,d0				* get key
	TRAP		#15

	BRA		start				* go get another file


*************************************************************************************
*
* setup stuff. only im_x and im_y are directly related to the image

Initialise
	MOVEQ		#0,d1				* turn off keyboard echo
	MOVEQ		#12,d0			* keyboard echo
	TRAP		#15

	MOVE.w	#$FF00,d1			* clear screen
	MOVEQ		#11,d0			* position cursor
	TRAP		#15

	MOVEQ		#50,d0			* close all files
	TRAP		#15

	LEA		variables(pc),a3		* set the variables pointer

	MOVE.w	#0,im_x(a3)			* set the image x origin
	MOVE.w	#0,im_y(a3)			* set the image y origin

	RTS


*************************************************************************************
*
* open a *.bmp file. returns the file id in d1.l or -1 if no file open

open_file
	MOVE.l	a3,-(sp)			* save the variables base pointer
	LEA		title(pc),a1		* set the request title string pointer
	LEA		list(pc),a2			* set the file types list pointer
	LEA		im_palette(a3),a3		* set the file name buffer pointer
	MOVEQ		#0,d1				* file open
	MOVE.b	d1,(a3)			* ensure null source file
	MOVEQ		#58,d0			* file I/O
	TRAP		#15

	MOVEA.l	a3,a1				* copy the file name pointer
	MOVEA.l	(sp)+,a3			* restore the variables pointer
	TST.l		d1				* did the user hit open
	BEQ.s		no_open_file		* if not do no file exit

	MOVE		#51,d0			* else open existing file
	TRAP		#15

	TST.w		d0				* test open result
	BEQ.s		exit_open_file		* if no error just exit

	MOVEQ		#50,d0			* else close all files
	TRAP		#15

no_open_file
	MOVEQ		#-1,d1			* clear the file id
exit_open_file
	MOVE.l	d1,file_id(a3)		* save the file ID
	RTS

title
	dc.b	'Open bitmap file',0
list
	dc.b	'*.bmp',0				* file type list
	ds.w	0					* ensure even


*************************************************************************************
*
* load the file. this first loads the header and checks for the "BM" signature. if
* the load succeedes and the signature is correct it loads the body of the file and
* checks that it too has loaded from the expected byte count
*
* returns d1.l = 0 if successful, d1.l = -1 if failed

* the bitmap file header is made up as follows
*
*	dc.b	'BM'					* *.bmp file header signature bytes
*	dc.l	size					* the total size of the file
*	dc.w	reserved				* usually zero
*	dc.w	reserved				* usually zero
*	dc.l	offset				* the offset to the start of the image data
*							* from the start of the file

load_file
	MOVE.l	file_id(a3),d1		* get the file ID

	LEA		im_file(a3),a1		* point to the file buffer
	MOVEQ		#$0E,d2			* set the byte count to load the file header
	MOVE.l	d2,d3				* copy the file header byte count
	MOVEQ		#53,d0			* read from the file
	TRAP		#15

	CMP.l		d3,d2				* compare the bytes read with the expected
							* byte count
	BNE.s		exit_load_file		* if incorrect do failure exit

	CMPI.w	#'BM',(a1)+			* compare the signature word with "BM"
	BNE.s		exit_load_file		* if incorrect do failure exit

	NEG.l		d2				* negate the byte count so far
	BSR		get_long			* get the file size
	ADD.l		d0,d2				* add the file size longword to give the
							* remaining byte count
	MOVE.l	d2,d3				* copy the remaining byte count
	ADDQ.w	#4,a1				* skip the two reserved words
	BSR		get_long			* get the image start offset
	SUB.l		#$0E,d0			* subtract the file header size
	MOVE.l	d0,image_start(a3)	* save the image start offset

	LEA		im_file(a3),a1		* point to the file buffer
	MOVEQ		#53,d0			* read from the file
	TRAP		#15

	CMP.l		d3,d2				* compare the bytes read with the expected
							* byte count
	BNE.s		exit_load_file		* if incorrect do failure exit

	MOVEQ		#0,d1				* flag load succeeded
	RTS

exit_load_file
	MOVEQ		#-1,d1			* flag load failed
	RTS


*************************************************************************************
*
* check the *.bmp image header and get the image information, if needed load the
* palette

* the bitmap image header is made up as follows
*
*	dc.l	header_size				* should be 40
*	dc.l	im_width				* image width in pixels
*	dc.l	im_height				* image height in pixels
*	dc.w	bitplanes				* number of planes, should be 1
*	dc.w	im_depth				* number of bits per pixel, should be one
*							* of 1, 4 8 or 24
*	dc.l	compression				* type of compression, should be zero, no
*							* compression
*	dc.l	im_size				* size of the image data in bytes. if there
*							* is no compression this can be set to zero
*	dc.l	im_x_ppm				* x resolution in pixels per meter, usually zero
*	dc.l	im_y_ppm				* y resolution in pixels per meter, usually zero
*	dc.l	colours_used			* number of colours used, usually zero
*	dc.l	colours_important			* number of important colours, usually zero

bm_image_header
	MOVEQ		#-1,d1			* flag failed

	MOVEQ		#$28,d2			* set the size that is expected
	BSR		get_long			* get the image header size
	CMP.l		d2,d0				* compare them
	BNE.s		exit_bm_image_header	* if incorrect just exit

	BSR		get_long			* get the image width
	MOVE.l	d0,im_width(a3)		* save the image width
	BSR		get_long			* get the image height
	MOVE.l	d0,im_height(a3)		* save the image height

	BSR		get_word			* get the number of planes
	CMPI.w	#1,d0				* compare with expected number
	BNE.s		exit_bm_image_header	* if incorrect just exit

	BSR		get_word			* get the image depth
	MOVE.w	d0,im_depth(a3)		* save the image depth

	BSR		get_long			* get the compression type
	MOVE.l	d0,d1				* flag success only if d0 is zero
	BNE.s		exit_bm_image_header	* if incorrect just exit

	LEA		$14(a1),a1			* skip the rest of the header

* load the palette. if the palette is anything other than 2, 16 or 256 colours this
* part of the routine does not loading anything

load_palette
	MOVE.w	im_depth(a3),d0		* get the image depth
	CMPI.w	#8,d0				* compare the image depth with 256 colour
	BGT.s		exit_load_palette		* if > 8 bpp just exit

	SUBQ.w	#1,d0				* make 1 to 8 into 0 to 7
	BMI.s		exit_load_palette		* if < 1 bpp just exit

	ADD.w		d0,d0				* two bytes per value
	MOVE.w	bpp_table(pc,d0.w),d7	* get the number of colours - 1
	BMI.s		exit_load_palette		* if not a valid depth just exit

	LEA		im_palette(a3),a2		* point to the palette

* get the bitmap palette colour values. palette colour values are 32 bit values stored
* as B,G,R,$00. EASy68K wants 32 bit colour values as $00,B,G,R

load_palette_loop
	MOVEQ		#0,d0				* clear the longword
	MOVE.b	(a1)+,d0			* get the blue byte
	SWAP		d0				* move it to the high word
	MOVE.b	(a1)+,d0			* get the green byte
	ROL.w		#8,d0				* move it to the high byte
	MOVE.b	(a1)+,d0			* get the red byte
	ADDQ.w	#1,a1				* skip the zero byte

	MOVE.l	d0,(a2)+			* save the value to the palette

	DBF		d7,load_palette_loop	* loop for the next colour

exit_load_palette

exit_bm_image_header
	RTS

bpp_table
	dc.w		$0001,-1,-1,$000F,-1,-1,-1,$00FF


*************************************************************************************
*
* display the image. if the image is smaller than the window set the x and y offsets
* so that the image is centred

display_image
	MOVEQ		#0,d1				* get the window size
	MOVEQ 	#33,d0			* get/set the window size
	TRAP		#15

	SUB.w		im_height+2(a3),d1	* subtract the image height from the window
							* height
	BMI.s		do_width			* if greater skip setting the y offset

	ASR.w		#1,d1				* / 2 to centre the inage
	MOVE.w	d1,im_y(a3)			* set the image y origin
do_width
	SWAP		d1				* shift the image width to the low word
	SUB.w		im_width+2(a3),d1		* subtract the image width from the window
							* width
	BMI.s		do_image			* if greater skip setting the x offset

	ASR.w		#1,d1				* / 2 to centre the inage
	MOVE.w	d1,im_x(a3)			* set the image x origin
do_image
	SUBQ.l	#1,im_width(a3)		* make the image width 0 to n

	MOVE.w	im_y(a3),d5			* get the image y origin
	MOVE.l	im_height(a3),d7		* get the image height
	SUBQ.l	#1,d7				* adjust for the loop type
	ADD.w		d7,d5				* set the end y co-ordinate

	LEA		im_file(a3),a1		* point to the file buffer
	ADDA.l	image_start(a3),a1	* add the image start offset

	MOVE.w	im_depth(a3),d0		* get the image depth
	CMP.w		#1,d0				* compare with one bit per pixel
	BEQ.s		display_1bbp		* if 1bbp go do 1bpp display

	CMP.w		#4,d0				* compare with four bits per pixel
	BEQ		display_4bbp		* if 4bbp go do 4bpp display

	CMP.w		#8,d0				* compare with eight bits per pixel
	BEQ		display_8bbp		* if 8bbp go do 8bpp display

	CMP.w		#24,d0			* compare with twenty four bits per pixel
	BEQ		display_real		* if 24bbp go do real colour display

	RTS


*************************************************************************************
*
* load a one bit, two colour, bitmap. image line data is padded to the next even word
* so unused bits and bytes have to be skipped

display_1bbp
	MOVE.l	im_width(a3),d6		* get the image width
	ADDQ.l	#1,d6				* make 1 to n
	NEG.l		d6				* make -ve
	MOVEQ		#24,d3			* set the byte remainder mask
	AND.l		d6,d3				* mask the byte remainder
	ASR.w		#3,d3				* / 8 pixels per byte
bpp1_y_loop
	MOVEM.l	d3/d7,-(sp)			* save the y count and byte remainder

	MOVE.l	im_width(a3),d6		* get the image width
	MOVE.w	im_x(a3),d4			* get the image x origin
bbp1_byte_loop
	MOVE.b	(a1)+,d7			* get a pixel word

	MOVEQ		#$07,d3			* set the bit count
bbp1_pixel_loop
	ROL.b		#1,d7				* shift the pixels
	MOVEQ		#1,d1				* set the pixel mask
	AND.b		d7,d1				* get the pixel
	BSR		get_colour			* get the pen colour and draw the pixel

	SUBQ.l	#1,d6				* decrement the pixel count
	BMI.s		bbp1_next_line		* exit the pixel loop if done

	DBF		d3,bbp1_pixel_loop	* loop for next pixel

	BRA.s		bbp1_byte_loop		* loop for next byte

bbp1_next_line
	MOVEM.l	(sp)+,d3/d7			* restore the y count and byte remainder
	ADDA.w	d3,a1				* skip the remainder bytes
	SUBQ.w	#1,d5				* decrement the y co-ordinate
	DBF		d7,bpp1_y_loop		* loop for next line

	RTS


*************************************************************************************
*
* load a four bit, sixteen colour, bitmap. image line data is padded to the next even
* word so unused nibbles and bytes have to be skipped

display_4bbp
	MOVE.l	im_width(a3),d6		* get the image width
	ADDQ.l	#1,d6				* make 1 to n
	NEG.l		d6				* make -ve
	MOVEQ		#6,d3				* set the byte remainder mask
	AND.l		d6,d3				* mask the byte remainder
	ASR.w		#1,d3				* / 2 pixels per byte
bpp4_y_loop
	MOVE.l	d7,-(sp)			* save the y count

	MOVE.l	im_width(a3),d6		* get the image width
	MOVE.w	im_x(a3),d4			* get the image x origin
bbp4_byte_loop
	MOVE.b	(a1)+,d7			* get a pixel word
	ROL.b		#4,d7				* shift the pixels
	MOVEQ		#$0F,d1			* set the pixel mask
	AND.w		d7,d1				* get the pixel
	BSR		get_colour			* get the pen colour and draw the pixel

	SUBQ.l	#1,d6				* decrement the pixel count
	BMI.s		bbp4_next_line		* exit the pixel loop if done

	ROL.b		#4,d7				* shift the pixels
	MOVEQ		#$0F,d1			* set the pixel mask
	AND.w		d7,d1				* get the pixel
	BSR		get_colour			* get the pen colour and draw the pixel

	DBF		d6,bbp4_byte_loop		* loop for next line

bbp4_next_line
	ADDA.w	d3,a1				* skip the remainder bytes
	SUBQ.w	#1,d5				* decrement the y co-ordinate
	MOVE.l	(sp)+,d7			* restore the y count
	DBF		d7,bpp4_y_loop		* loop for next line

	RTS


*************************************************************************************
*
* load an eight bit, two hundred and fifty six colour, bitmap. image line data is
* padded to the next even word so unused bytes have to be skipped

display_8bbp
	MOVE.l	im_width(a3),d6		* get the image width
	ADDQ.l	#1,d6				* make 1 to n
	NEG.l		d6				* make -ve
	MOVEQ		#3,d3				* set the byte remainder mask
	AND.l		d6,d3				* mask the byte remainder
bpp8_y_loop
	MOVE.l	im_width(a3),d6		* get the image width
	MOVE.w	im_x(a3),d4			* get the image x origin
bbp8_x_loop
	MOVEQ		#0,d1				* clear the longword
	MOVE.b	(a1)+,d1			* get the pixel
	BSR		get_colour			* get the pen colour and draw the pixel

	DBF		d6,bbp8_x_loop		* loop for the next pixel

	ADDA.w	d3,a1				* skip the remainder bytes
	SUBQ.w	#1,d5				* decrement the y co-ordinate
	DBF		d7,bpp8_y_loop		* loop for next line

	RTS


*************************************************************************************
*
* load a twenty four bit, real colour, bitmap. image line data is padded to the next
* even word so unused bytes have to be skipped

display_real
	MOVE.l	im_width(a3),d6		* get the image width
	ADDQ.l	#1,d6				* make 1 to n
	MOVEQ		#3,d3				* set byte remainder mask
	AND.l		d6,d3				* mask byte remainder
real_y_loop
	MOVE.l	im_width(a3),d6		* get the image width
	MOVE.w	im_x(a3),d4			* get the image x origin
real_x_loop
	MOVEQ		#0,d1				* clear the longword
	MOVE.b	(a1)+,d1			* get the blue byte
	SWAP		d1				* move it to the high word
	MOVE.b	(a1)+,d1			* get the green byte
	ROL.w		#8,d1				* move it to the high byte
	MOVE.b	(a1)+,d1			* get the red byte

	BSR		set_colour			* set the pen colour and draw the pixel

	DBF		d6,real_x_loop		* loop for the next pixel

	ADDA.w	d3,a1				* skip the remainder bytes
	SUBQ.w	#1,d5				* decrement the y co-ordinate
	DBF		d7,real_y_loop		* loop for next line

	RTS


*************************************************************************************
*
* get the pen colour from the palette colour indexed by d1.w, set the colour and draw
* the pixel

get_colour
	ASL.w		#2,d1				* 4 bytes per longword colour
	MOVE.l	im_palette(a3,d1.w),d1	* get the colour

*------------------------------------------------------------------------------------
*
* set the pen colour from d1.l and draw the pixel

set_colour
	MOVEQ		#80,d0			* set the pen colour
	TRAP		#15

	MOVE.w	d4,d1				* copy the x co-ordinate
	MOVE.w	d5,d2				* copy the y co-ordinate

	MOVEQ		#82,d0			* draw a pixel
	TRAP		#15

	ADDQ.w	#1,d4				* increment the x co-ordinate
	RTS


*************************************************************************************
*
* get a little endian word. returns the result in d0.w and increments the pointer

get_word
	MOVE.b	(a1)+,d0			* get the low byte
	ROL.w		#8,d0				* move it
	MOVE.b	(a1)+,d0			* get the high byte
	ROL.w		#8,d0				* swap the bytes
	RTS


*************************************************************************************
*
* get a little endian longword. returns the result in d0.l and increments the pointer

get_long
	BSR.s		get_word			* get a little endian word
	SWAP		d0				* move it
	BSR.s		get_word			* get a little endian word
	SWAP		d0				* swap the words
	RTS


*************************************************************************************
*
* variables

variables						* variables base address

file_id					EQU	*-variables
	ds.l	1					* file ID longword

image_start					EQU	*-variables
	ds.l	1					* offset to the start of the image data

im_x						EQU	*-variables
	ds.w	1					* image x origin
im_y						EQU	*-variables
	ds.w	1					* image y origin

im_width					EQU	*-variables
	ds.l	1					* image width
im_height					EQU	*-variables
	ds.l	1					* image height
im_depth					EQU	*-variables
	ds.w	1					* image depth in bits per pixel
im_palette					EQU	*-variables
	ds.l	256					* image palette space if needed

im_file					EQU	*-variables
							* the rest of the memory is the file buffer


	END	start


*************************************************************************************